package eu.jonahbauer.raytracing.render.material;

import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.render.renderer.pdf.ProbabilityDensityFunction;
import eu.jonahbauer.raytracing.render.texture.Color;
import eu.jonahbauer.raytracing.render.texture.Texture;
import eu.jonahbauer.raytracing.scene.HitResult;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.Optional;
import java.util.random.RandomGenerator;

public interface Material {

    /**
     * {@return the texture associated with this material}
     */
    @NotNull Texture texture();

    /**
     * Scatters a light ray after it hit a surface.
     * @param ray the incoming light ray
     * @param hit information about the light ray hitting some object
     * @param random a random number generator
     * @return a {@code ScatterResult} if the ray is scattered or an {@linkplain Optional#empty() empty optional} if the
     * ray is absorbed.
     */
    @NotNull Optional<ScatterResult> scatter(@NotNull Ray ray, @NotNull HitResult hit, @NotNull RandomGenerator random);

    /**
     * {@return the color emitted for a given hit}
     * @implSpec the default implementation returns {@linkplain Color#BLACK black}, i.e. no emission
     */
    default @NotNull Color emitted(@NotNull HitResult hit) {
        return Color.BLACK;
    }

    /**
     * The result of a {@linkplain Material#scatter(Ray, HitResult, RandomGenerator) scattering operation}.
     */
    sealed interface ScatterResult {}

    /**
     * The result of a specular {@linkplain #scatter(Ray, HitResult, RandomGenerator) scattering operation}. A
     * specular is a scattering operation with a very small number of possible scattered rays (like a
     * perfect reflection which only has one possible scattered ray).
     * @param attenuation the attenuation of the scattered light ray
     * @param ray the scattered light ray
     */
    record SpecularScatterResult(@NotNull Color attenuation, @NotNull Ray ray) implements ScatterResult {
        public SpecularScatterResult {
            Objects.requireNonNull(attenuation, "attenuation");
            Objects.requireNonNull(ray, "ray");
        }
    }

    /**
     * The result of a probability density function based
     * {@linkplain #scatter(Ray, HitResult, RandomGenerator) scattering operation}. A probability density function
     * based scattering operation uses a probability density function to determine the scatter direction.
     * @param attenuation the attenuation of the scattered light ray
     * @param pdf the probability density function
     */
    record PdfScatterResult(@NotNull Color attenuation, @NotNull ProbabilityDensityFunction pdf) implements ScatterResult {
        public PdfScatterResult {
            Objects.requireNonNull(attenuation, "attenuation");
            Objects.requireNonNull(pdf, "pdf");
        }
    }
}
